package cn.iocoder.yudao.module.bpm.service.oa;

import cn.iocoder.yudao.framework.common.pojo.PageResult;
import cn.iocoder.yudao.framework.common.pojo.UploadUserFile;
import cn.iocoder.yudao.framework.common.util.object.BeanUtils;
import cn.iocoder.yudao.framework.mybatis.core.util.MyBatisUtils;
import cn.iocoder.yudao.module.bpm.api.task.BpmProcessInstanceApi;
import cn.iocoder.yudao.module.bpm.api.task.dto.BpmProcessInstanceCreateReqDTO;
import cn.iocoder.yudao.module.bpm.controller.admin.oa.vo.expenses.BpmOAExpensesCreateReqVO;
import cn.iocoder.yudao.module.bpm.controller.admin.oa.vo.expenses.BpmOAExpensesPageReqVO;
import cn.iocoder.yudao.module.bpm.controller.admin.oa.vo.expenses.BpmOAExpensesPageRespVO;
import cn.iocoder.yudao.module.bpm.controller.admin.oa.vo.expenses.BpmOAExpensesRespVO;
import cn.iocoder.yudao.module.bpm.convert.oa.BpmOAExpensesConvert;
import cn.iocoder.yudao.module.bpm.dal.dataobject.oa.BpmOAExpensesDO;
import cn.iocoder.yudao.module.bpm.dal.dataobject.oa.BpmOAExpensesItemDO;
import cn.iocoder.yudao.module.bpm.dal.mysql.oa.BpmOAExpensesItemMapper;
import cn.iocoder.yudao.module.bpm.dal.mysql.oa.BpmOAExpensesMapper;
import cn.iocoder.yudao.module.bpm.enums.task.BpmProcessInstanceResultEnum;
import cn.iocoder.yudao.module.bpm.service.task.BpmHistoryProcessInstanceService;
import cn.iocoder.yudao.module.system.api.bank.BankApi;
import cn.iocoder.yudao.module.system.api.bank.dto.BankRespDTO;
import cn.iocoder.yudao.module.system.api.dept.DeptApi;
import cn.iocoder.yudao.module.system.api.dept.dto.DeptRespDTO;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.annotation.Validated;

import javax.annotation.Resource;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.framework.common.util.collection.CollectionUtils.convertMap;
import static cn.iocoder.yudao.framework.common.util.collection.CollectionUtils.convertSet;
import static cn.iocoder.yudao.module.bpm.enums.ErrorCodeConstants.OA_EXPENSES_NOT_EXISTS;

/**
 * OA 生产开支 Service 实现类
 *
 * @author 符溶馨
 */
@Service
@Validated
public class BpmOAExpensesServiceImpl extends BpmOABaseService implements BpmOAExpensesService {

    /**
     * OA 生产开支对应的流程定义 KEY
     */
    public static final String PROCESS_KEY = "oa_expenses_2";

    @Resource
    private BpmOAExpensesMapper expensesMapper;

    @Resource
    private BpmOAExpensesItemMapper expensesItemMapper;

    @Resource
    private BpmProcessInstanceApi processInstanceApi;

    @Resource
    private DeptApi deptApi;

    @Resource
    private BankApi bankApi;

    @Resource
    @Lazy // 解决循环依赖
    private BpmHistoryProcessInstanceService historyProcessInstanceService;

    @Override
    public Long createExpenses(Long userId, BpmOAExpensesCreateReqVO createReqVO) {

        //插入OA 生产开支申请
        BpmOAExpensesDO expenses = BpmOAExpensesConvert.INSTANCE.convert(createReqVO)
                .setUserId(userId)
                .setResult(BpmProcessInstanceResultEnum.PROCESS.getResult());
        expensesMapper.insert(expenses);

        // 批量插入生产开支明细数据
        List<BpmOAExpensesItemDO> bpmOAExpensesItemDOS = BeanUtils.toBean(createReqVO.getExpensesItem(), BpmOAExpensesItemDO.class);
        bpmOAExpensesItemDOS.forEach(item -> item.setExpensesId(expenses.getId()));
        expensesItemMapper.insertBatch(bpmOAExpensesItemDOS);

        // 发起 BPM 流程
        Map<String, Object> processInstanceVariables = new HashMap<>();
        String type = bpmOAExpensesItemDOS.stream().map(item -> item.getType().toString()).collect(Collectors.joining(","));
        processInstanceVariables.put("type", type);
        String processInstanceId = processInstanceApi.createProcessInstance(userId,
                new BpmProcessInstanceCreateReqDTO().setProcessDefinitionKey(PROCESS_KEY)
                        .setVariables(processInstanceVariables).setBusinessKey(String.valueOf(expenses.getId()))).getCheckedData();

        // 将工作流的编号，更新到 OA 生产开支单中
        expensesMapper.updateById(new BpmOAExpensesDO().setId(expenses.getId()).setProcessInstanceId(processInstanceId));

        // 判断是否为重新发起的流程
        if (createReqVO.getProcessInstanceId() != null && createReqVO.getResult() == 3) {

            historyProcessInstanceService.createHistoryProcessInstance(processInstanceId, createReqVO.getProcessInstanceId());
        }

        List<UploadUserFile> fileItems = createReqVO.getFileItems();
        //这里的逻辑，如果fileItems不为空，且有数据，那么说明是上传了附件的，则需要更工作流文件表对应的实例Id
        if (fileItems != null && !fileItems.isEmpty()) {
            uploadBpmFileProcessInstanceId(processInstanceId, fileItems);
        }
        return expenses.getId();
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void updateExpensesResult(String processInstanceId, Long id, Integer result) {

        // 获得现金支出业务数据
        BpmOAExpensesDO expenses = validateLeaveExists(id);

        expensesMapper.updateById(new BpmOAExpensesDO().setId(id).setResult(result));
    }

    private BpmOAExpensesDO validateLeaveExists(Long id) {
        BpmOAExpensesDO expenses = expensesMapper.selectById(id);
        if (expenses == null) {
            throw exception(OA_EXPENSES_NOT_EXISTS);
        }

        return expenses;
    }

    @Override
    public BpmOAExpensesDO getExpenses(Long id) {

        return expensesMapper.selectById(id);
    }

    @Override
    public List<BpmOAExpensesItemDO> getExpensesItem(Long expensesId) {

        return expensesItemMapper.selectList(BpmOAExpensesItemDO::getExpensesId, expensesId);
    }

    @Override
    public BpmOAExpensesDO getByProcessInstanceId(String processInstanceId) {
        List<BpmOAExpensesDO> list = expensesMapper.selectList(
                new LambdaQueryWrapper<BpmOAExpensesDO>()
                        .eq(BpmOAExpensesDO::getProcessInstanceId, processInstanceId));
        if (!list.isEmpty()) {
            return list.get(0);
        }
        return null;
    }

    @Override
    public BpmOAExpensesRespVO convertExpenses(BpmOAExpensesDO expenses) {

        // 获取明细数据
        List<BpmOAExpensesItemDO> expensesItemDOs = getExpensesItem(expenses.getId());

        //获取部门信息map
        List<DeptRespDTO> deptList = deptApi.getDeptByFactoryIds(convertSet(expensesItemDOs, BpmOAExpensesItemDO::getDeptId)).getCheckedData();
        Map<Long, DeptRespDTO> deptMap = convertMap(deptList, DeptRespDTO::getFactoryId);

        // 获取银行卡信息
        BankRespDTO bankRespDTO = null;
        if (expenses.getBankId() != null) {
            bankRespDTO = bankApi.getBank(expenses.getBankId()).getCheckedData();
        }

        return BpmOAExpensesConvert.INSTANCE.convert1(expenses, expensesItemDOs, deptMap, bankRespDTO);
    }

    @Override
    public PageResult<BpmOAExpensesPageRespVO> getExpensesPage(BpmOAExpensesPageReqVO pageReqVO) {
        IPage<BpmOAExpensesPageRespVO> page = expensesMapper.selectExpensesPage(MyBatisUtils.buildPage(pageReqVO) ,pageReqVO);
        return new PageResult<>(page.getRecords(), page.getTotal());
    }
}
